










/* gvfs.c -- The virtual filesystem layer interface.

 * Copyright (C) 1999 Steve Hill <sjhill@plutonium.net>
 *                    Yarick Rastrigin <yarick@relex.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <dirent.h>
#include <stdio.h>
#include <strings.h>
#include <errno.h>

#include <glib.h>
#include <gtk/gtk.h>

#include "common.h"
#ifdef __DEBUG__
#include "debug.h"
#endif
#include "gvfs.h"

static GVFS *gvfs_handlers = &gvfs_local_fs_ops;
static gchar *curr_dir = NULL;

typedef struct _TempSortStruct
  {
    gchar *name;
    void *info;
  }
TempSortStruct;


gint
GvfsListCompareFunc (void *a, void *b)
{
  /* Temp wrapper for future diff.sorting methods. */
  return strcmp (((TempSortStruct *) (a))->name, ((TempSortStruct *) (b))->name);
}

gint
gvfs_register (GVFS * gvfs)
{
  int res;

  /* Check for NULL filesystem handler. */
  if (!gvfs)
    {
#ifdef __DEBUG__
      debug_out ("Attempted registering of NULL virtual filesystem handler.");
#endif
      exit (1);
    }

  /* If the filesystem handler has an initialization function, go ahead
   * and execute it.
   */
  res = (gvfs->init) ? (*gvfs->init) (gvfs) : 1;

  /* Check the return value of the filesystem handler's initialization
   * function. If it did not initialize successfully, do not add it
   * to the list of valid virtual filesystem handlers.
   */
  if (!res)
    return 0;

  /* Add the handler to the list and return succesfully. */
  gvfs->next = gvfs_handlers;
  gvfs_handlers = gvfs;

  return 1;
}

guint
gvfs_initialize (void)
{
  /* Set current directory. */
  curr_dir = (gchar *) g_malloc0 (PATH_MAX * sizeof (gchar));
  getcwd (curr_dir, PATH_MAX);

  /* Register the virtual filesystem layer handlers. */
  gvfs_register (&gvfs_smb_fs_ops);

  return 1;
}

void
gvfs_shutdown (void)
{
  g_free (curr_dir);
//  g_free (gvfs_handlers);
}

gchar *
gvfs_get_current_dir (void)
{
  return g_strdup (curr_dir);
}


void
gvfs_set_current_dir (gchar * directory)
{
  strncpy (curr_dir, directory, PATH_MAX);
}


gboolean
gvfs_has_subdirs (gchar * path)
{
  DIR *dir;
  struct dirent *dir_entry;
  struct stat dir_entry_status;
  gboolean exists;
  gchar *working_path;
  GVFS *gvfs;

  /* Allocate space for path and initialize value. */
  working_path = (gchar *) g_malloc0 (PATH_MAX * sizeof (gchar));
  exists = FALSE;

  /* Get the proper GVFS handler. */
  gvfs = gvfs_type (path);
  if (gvfs == NULL)
    {
      debug_out ("NULL handler in 'gvfs_has_subdirs'.");
      return exists;
    }

  /* Open the directory for reading. */
  dir = (*gvfs->opendir) (path);
  if (dir != NULL)
    {
      while ((dir_entry = (*gvfs->readdir) (dir)) != NULL)
	{
	  /* Exclude the "." and ".." directories. */
	  if ((strcmp (dir_entry->d_name, ".") != 0) &&
	      (strcmp (dir_entry->d_name, "..") != 0))
	    {
	      /* Create the absolute path. */
	      strcpy (working_path, path);
	      strcat (working_path, "/");
	      strcat (working_path, dir_entry->d_name);

	      /* See if directory entry is a directory. */
	      (*gvfs->stat) (working_path, &dir_entry_status);
	      if (S_ISDIR (dir_entry_status.st_mode))
		exists = TRUE;
	    }
	}

      /* Close the directory. */
      (*gvfs->closedir) (dir);
    }

  /* Return the value. */
  return exists;
}

GList *
gvfs_get_x_list (gchar * path, FSItemTypeCheckFunc check_func, gboolean need_sort, GList ** stat_info)
{
  DIR *dir;
  struct dirent *dir_entry;
  struct stat dir_entry_status;
  GList *dir_list, *hGL1, *hGL2, *hGL3;
  TempSortStruct *hTSS1;
  gchar *working_path, *str1, *str2, *str3;
  GVFS *gvfs;
  gpointer stat_data;

  /* Allocate space for path. */
  working_path = (gchar *) g_malloc0 (PATH_MAX * sizeof (gchar));

  /* Get the proper GVFS handler. */
  gvfs = gvfs_type (path);
  if (gvfs == NULL)
    {
#ifdef __DEBUG__
      debug_out ("NULL handler in 'gvfs_get_subdirs'.");
#endif
      return NULL;
    }

  /* Open the directory for reading. */
  dir_list = NULL;
  dir = (*gvfs->opendir) (path);
  if (dir != NULL)
    {
      while ((dir_entry = (*gvfs->readdir) (dir)) != NULL)
	{
	  /* Exclude the "." and ".." directories. */
	  if ((strcmp (dir_entry->d_name, ".") != 0) &&
	      (strcmp (dir_entry->d_name, "..") != 0))
	    {
	      /* Create the absolute path. */
	      strcpy (working_path, path);
	      if (*(working_path + strlen (working_path) - 1) != '/')
		strcat (working_path, "/");
	      strcat (working_path, dir_entry->d_name);

	      /* See if directory entry is a directory. */
	      (*gvfs->lstat) (working_path, &dir_entry_status);
	      if (S_ISLNK (dir_entry_status.st_mode))
		{
		  (*gvfs->stat) (working_path, &dir_entry_status);
		  dir_entry_status.st_nlink |= 0100000;
		  /* we're signaling about symbolic link through high bit set in st_nlink
		     Rather strange method, but adding fields would probably break things up for
		     a long time.
		   */
		};
	      if (check_func (&dir_entry_status))
		{
		  dir_list = g_list_append (dir_list, g_strdup (dir_entry->d_name));
		  if (stat_info != NULL)
		    {
		      stat_data = g_malloc (sizeof (struct stat));
		      memcpy (stat_data, &dir_entry_status, sizeof (struct stat));
		      *stat_info = g_list_append (*stat_info, stat_data);
		    };
		};
	    };
	}

      /* Close the directory. */
      (*gvfs->closedir) (dir);
    }
  else
    return NULL;

  /* Sort the subdirectory list. */
  if (dir_list == NULL)
    return NULL;

  if (!need_sort)
    return dir_list;

/* Building temp sort list and helper lists */
  hGL1 = NULL;
  if (stat_info != NULL)
    hGL3 = *stat_info;
  else
    hGL3 = NULL;
  for (hGL2 = dir_list; hGL2 != NULL; hGL2 = g_list_next (hGL2))
    {
      hTSS1 = g_malloc0 (sizeof (TempSortStruct));
      hTSS1->name = hGL2->data;
      if (hGL3 != NULL)
	{
	  hTSS1->info = hGL3->data;
	  hGL3 = hGL3->next;
	};
      hGL1 = g_list_append (hGL1, hTSS1);
    };
  /* Soritng ... */
  hGL1 = g_list_sort (hGL1, (GCompareFunc) GvfsListCompareFunc);
  str1 = (gchar *) hGL1;
/*  And recovering ... */
  if (stat_info != NULL)
    hGL3 = *stat_info;
  else
    hGL3 = NULL;
  for (hGL2 = dir_list; hGL2 != NULL; hGL2 = g_list_next (hGL2))
    {
      hTSS1 = hGL1->data;
      hGL2->data = hTSS1->name;
      if (hGL3 != NULL)
	{
	  hGL3->data = hTSS1->info;
	  hGL3 = g_list_next (hGL3);
	};
      g_free (hTSS1);
      hGL1 = g_list_next (hGL1);
    };
  hGL1 = (GList *) str1;
  g_list_free (hGL1);
  /* I just don't want to create another variable for storing list's first element */
  /* Return the list. */
  return dir_list;
}

GVFS *
gvfs_type (gchar * path)
{
  GVFS *gvfs;
  gchar *gvfs_type;
  guint length;

  /* Set return pointer to NULL. */
  gvfs = NULL;

  /* Check to see if path is on the local filesystem structure. */
  if (*path == '/')
    {
      gvfs = &gvfs_local_fs_ops;
      return gvfs;
    }

  /* Check to see what kind of filesystem we have. */
  length = strcspn (path, ":");
  if (length < 1)
    return gvfs;
  gvfs_type = (gchar *) g_malloc0 ((length + 1) * sizeof (gchar));
  strncpy (gvfs_type, path, length);

  /* Check to see if path is a FTP site. */
  if (strcmp (gvfs_type, "ftp") == 0)
    gvfs = &gvfs_ftp_fs_ops;

  /* Check to see if path is a SMB share. */
  if (strcmp (gvfs_type, "smb") == 0)
    gvfs = &gvfs_smb_fs_ops;

  /* Free memory and return GVFS pointer. */
  g_free (gvfs_type);
  return gvfs;
}

void *
gvfs_opendir (GVFS * gvfs, gchar * dirname)
{
  return (*gvfs->opendir) (dirname);
}

void *
gvfs_readdir (GVFS * gvfs, void *dir_info)
{
  return (*gvfs->readdir) (dir_info);
}

gint
gvfs_closedir (GVFS * gvfs, void *dir_info)
{
  return (*gvfs->closedir) (dir_info);
}

gint
gvfs_stat (GVFS * gvfs, gchar * filename, void *buf)
{
  return (*gvfs->stat) (filename, buf);
}

gint
gvfs_lstat (GVFS * gvfs, gchar * filename, void *buf)
{
  return (*gvfs->lstat) (filename, buf);
}

gint
gvfs_readlink (gchar * path, void *buf, gint bufsize)
{
  GVFS *gvfs = gvfs_type (path);
  return (*gvfs->readlink) (path, buf, bufsize);
}

gint
gvfs_unlink (gchar * filename)
{
  GVFS *gvfs;
  gvfs = gvfs_type (filename);
  return (*gvfs->unlink) (filename);
}

gint
gvfs_mkdir (gchar * filename)
{
  GVFS *gvfs;
  gint return_result = -1;
  gchar *full_path;
  if (filename != NULL)
    {
      full_path = absolutize_name (filename);
      gvfs = gvfs_type (full_path);
      if (gvfs != NULL)
	return_result = (*gvfs->mkdir) (full_path);
      else
	return_result = -1;
    };
  g_free (full_path);
  return return_result;
}

gint
gvfs_symlink (gchar * symlink_name, gchar * symlink_target)
{
  GVFS *n_gvfs, *t_gvfs;
  gint return_result = 0;
  if ((symlink_name == NULL) || (symlink_target == NULL))
    return -1;
  debug_out (symlink_name);
  debug_out (symlink_target);
  symlink_name = absolutize_name (symlink_name);
  n_gvfs = gvfs_type (symlink_name);
  t_gvfs = gvfs_type (symlink_target);
  if (n_gvfs != t_gvfs)
    {
      g_free (symlink_name);
      return -1;
    };
  return_result = (*n_gvfs->symlink) (symlink_name, symlink_target);
  g_free (symlink_name);
  return return_result;
}

gint
gvfs_rm (gchar * filename)
{
  GVFS *gvfs;
  gchar *full_path;
  gint return_result;
  if (filename == NULL)
    return -1;
  full_path = absolutize_name (filename);
  gvfs = gvfs_type (full_path);
  if (gvfs != NULL)
    return_result = (*gvfs->rm) (full_path);
  else
    return_result = -1;
  g_free (full_path);
  return return_result;

}

gint
gvfs_rmdir (gchar * filename)
{
  GVFS *gvfs;
  gchar *full_path;
  gint return_result;
  if (filename == NULL)
    return -1;
  full_path = absolutize_name (filename);
  gvfs = gvfs_type (full_path);
  if (gvfs != NULL)
    return_result = (*gvfs->rmdir) (full_path);
  else
    return_result = -1;
  g_free (full_path);
  return return_result;
}

gint
gvfs_rename (gchar * from, gchar * to)
{
  GVFS *f_gvfs, *t_gvfs;
  gint return_result;
  f_gvfs = gvfs_type (from);
  t_gvfs = gvfs_type (to);
  if (f_gvfs != t_gvfs)
    return -1;
  return ((*f_gvfs->mv) (from, to));
}

gint
gvfs_copy_file (gvfs_file_copy_struct * gfcs)
{
  gint return_result;
  if (gfcs->source_gvfs == NULL)
    {
      return_result = gvfs_init_fcs (gfcs);
      if (return_result != 0)
	{
	  gfcs->all_ok = FALSE;
	  gvfs_shut_fcs (gfcs);
	  return return_result;
	};
    };
  if (gfcs->done_bytes < gfcs->total_bytes)
    {
      return_result = (*gfcs->source_gvfs->fread) (gfcs->source_fd, gfcs->buffer, 4096);
      if (return_result < 0)
	{
	  gfcs->all_ok = FALSE;
	  gvfs_shut_fcs (gfcs);
	  return return_result;
	};
      return_result = (*gfcs->target_gvfs->fwrite) (gfcs->target_fd, gfcs->buffer, return_result);
      if (return_result < 0)
	{
	  gfcs->all_ok = FALSE;
	  gvfs_shut_fcs (gfcs);
	  return return_result;
	};
    };
  gfcs->done_bytes += return_result;
  if (gfcs->done_bytes >= gfcs->total_bytes)
    {
      if (gfcs->all_ok)
	(*gfcs->target_gvfs->chmod) (gfcs->target_name, gfcs->mode);
      gvfs_shut_fcs (gfcs);
    };
  return 0;
}

gint
gvfs_init_fcs (gvfs_file_copy_struct * gfcs)
{
  gfcs->buffer = g_malloc (4096);
  gfcs->source_gvfs = gvfs_type (gfcs->source_name);
  gfcs->target_gvfs = gvfs_type (gfcs->target_name);
  gfcs->source_fd = (*gfcs->source_gvfs->fopen) (gfcs->source_name, "rb");
  if (gfcs->source_fd == NULL)
    return errno;
  if (gfcs->force_overwrite == FALSE)
    {
      gfcs->target_fd = (*gfcs->target_gvfs->fopen) (gfcs->target_name, "r");
      if (gfcs->target_fd != NULL)
	return EEXIST;
    };
  gfcs->target_fd = (*gfcs->target_gvfs->fopen) (gfcs->target_name, "wb");
  if (gfcs->target_fd == NULL)
    return errno;
  return 0;
}

gint
gvfs_shut_fcs (gvfs_file_copy_struct * gfcs)
{
  if (gfcs->source_fd != NULL)
    (*gfcs->source_gvfs->fclose) (gfcs->source_fd);
  if (gfcs->target_fd != NULL)
    (*gfcs->target_gvfs->fclose) (gfcs->target_fd);
  if (gfcs->buffer != NULL)
    g_free (gfcs->buffer);
  gfcs->source_gvfs = NULL;
  gfcs->target_gvfs = NULL;
  return 0;
}


gint
gvfs_chmod (gchar * filename, int mode)
{
  return (*gvfs_type (filename)->chmod) (filename, mode);
}

gint
gvfs_chown (gchar * filename, int owner, int group)
{
  return (*gvfs_type (filename)->chown) (filename, owner, group);
}
